# -*- coding: utf-8 -*-
## Copyright (C) 2007 Ingeniweb

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING. If not, write to the
## Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import time

from App.config import getConfiguration
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.CMFPlone.utils import safe_hasattr
from Products.CMFCore.utils import UniqueObject, getToolByName

from Products.Archetypes.utils import DisplayList
from Products.Archetypes.ArchetypeTool import WidgetWrapper

from Products.DataGridField import DataGridField, DataGridWidget

from config import *
from Products.TagCloudExplorer import LOG

class ConfigletField:
    ''' A class that permits embedding of an archetype field in a tool
    for display its widget in configlet 
    be careful :
        to get the archetype field,
        you might want to call the field attribute of a ConfigletField instance,
        or CALL the instance, but not directly get it 
    '''
    security = ClassSecurityInfo()
    
    __allow_access_to_unprotected_subobjects__ = True
    
    def getValue(self):
        ''' returns the value of the field
        defined in the tool as 'fieldName' attribute '''
        tool = self.tool
        return getattr(tool, self.fieldName)
    
    def getAccessor(self, context):
        ''' redefines accessor of the field to adapt it for tools '''
        self.tool = getToolByName(context, self.tool_id)
        return self.getValue
    
    def __init__(self, field, tool_id):
        ''' field is an archetype field defined as in an ATContentType schema,
            tool_id is the id of the tool concerned by the configlet
        '''
        self.field = field
        self.field.getAccessor = self.getAccessor
        self.field.getEditAccessor = self.getAccessor
        #redefines accessor of the field to adapt it for tools 
        self.tool_id = tool_id
        self.fieldName = field.getName()
        self.widget = field.widget
    
    def __call__(self):
        ''' gets the archetype field '''
        return self.field


class TagCloudExplorerTool(PropertyManager, UniqueObject, SimpleItem):
    """
        Configuration and "engine" of Tag Cloud Explorer
        
    """
    id = SKIN_NAME + "_tool"
    tool_id = id
    portal_type = meta_type = PROJECT_NAME + "Tool"
    plone_tool = 1
    title = PROJECT_TITLE + " Tool"
    
    security = ClassSecurityInfo()
    
    _properties = (
         {
            "id" : "max_tags",
            "type" : "int",
            "mode" : "w",
            "label" : "Maximum Number of Tags (max_tags)",
        },
        {
            "id" : "min_tags_size",
            "type" : "int",
            "mode" : "w",
            "label" : "Minimum Tags Size (min_tags_size)",
        },
        {
            "id" : "max_tags_size",
            "type" : "int",
            "mode" : "w",
            "label" : "Maximum Tags Size (max_tags_size)",
        },
        {
            "id" : "scaling_method",
            "type" : "string",
            "mode" : "w",
            "label" : "Tags Scale Method (scaling_method)",
        },
        {
             "id": "allowed_states",
             "type": "lines",
             "label": "States of contents searched by tagcloud (allowed_states)",
             "mode": "w",
             }
        
    )
    
    # Init Properties
    max_tags = 20
    min_tags_size = 10
    max_tags_size = 16
    scaling_method = 'linear'
    scaled_sizes = {}
    allowed_states = ()
 
    # ##################################################
    # #    DATAGRIDFIELD WIDGET AND TAGS_SCALE FIELD   #
    # ##################################################
    tags_scale = ({'min':'', 'max':'', 'size':''},)
    
    tags_scale_field = ConfigletField(
     DataGridField('tags_scale',
         widget = DataGridWidget(
             i18n_domain="tag_cloud_explorer",                    
             label="Tags Scale mapping",
             label_msgid="label_tags_scale_field",
             description='''You can define, for some scales
                             (from 'min' occurences to 'max' occurences),
                             the size of the tags ''',
             description_msgid="help_tags_scale_field",
             ),
         allow_empty_rows=False,
         columns=('min', 'max', 'size')
                  ),
     tool_id)
    
    

    def _setScaledSizes(self):
        ''' [{'min': '1', 'max':'10', 'size': '8'},
            {'min': '11', 'max':'30', 'size': '9'}] ->
            {(1, 10): 8,
             (11, 30): 9}
             
             '''
        scaled_sizes = {}
        for scale in self.tags_scale:
            min = int(scale['min'] or 0)
            max = int(scale['max'] or 0)
            size = int(scale['size'] or 0)
            scaled_sizes[(min, max)] = size
        self._scaled_sizes = scaled_sizes
               
    def getScaledSizes(self):
        ''' gets mapping of sizes by occurence range
            {(1, 10): 8,
            (11, 30): 9}
        '''
        return self._scaled_sizes
    
    def getOccurenceScaledSize(self, occurence):
        ''' occurence : int '''
        occurence = int(occurence)
        for scale_min, scale_max in self._scaled_sizes.keys():
            if scale_min <= occurence <= scale_max:
                return self._scaled_sizes[(scale_min, scale_max)]
        return None
    
    def getAllowableStates(self):
        """ get selectable states """
        allStates = getToolByName(self, 'portal_workflow').listWFStatesByTitle()
        return list(set(allStates))
    
    # ################################################
    # #         UPDATE PROPERTIES                    #
    # ################################################
    
    security.declareProtected(CONFIG_PERMISSION, "updateProperties")
    def updateProperties(self, 
                         min_tags_size=0,
                         max_tags_size=0,
                         max_tags=0,
                         tags_scale=[{}],
                         allowed_states=[],
                         scaling_method='linear'):
        ''' update tag cloud preferences as tool properties '''
        self.scaling_method = str(scaling_method)
        self.max_tags = int(max_tags)
        self.min_tags_size = int(min_tags_size)
        self.max_tags_size = int(max_tags_size)
        #tags_scale = self.getNonEmptyRows(tags_scale)
        self.allowed_states = allowed_states
        self.tags_scale_field().set(self, value=tags_scale)
        self._setScaledSizes()
      
      
    def __init__(self):
        self._setScaledSizes()      

    def getSearchClass(self, tag, REQUEST=None):
        """ Return css class for keywords in the search page """
        REQUEST = REQUEST or self.REQUEST
        subjects = REQUEST.get('subjects',[])
        if isinstance(subjects, str):
            subjects = [subjects]

        if tag in subjects:
            return 'tagcloud_selected'
        return 'tagcloud_unselected'
    
    def getPageKeywords(self, page_object):
        """ Return the current page's tags, alpha sorted """
        
        tags = list(page_object.Subject())
        if tags:
            tags.sort(lambda x, y: cmp(x.lower(), y.lower()))
            return tuple(tags)
    
    def getSiteKeywords(self):
        """ Return the website tags, limited by max_tag attribute
            Weight selected and/then alpha sorted                 """
            
        catalog = getToolByName(self, 'portal_catalog')
        utils   = getToolByName(self, 'plone_utils')
        tags    = list(catalog.uniqueValuesFor('Subject'))
        friendly_types = utils.getUserFriendlyTypes()
        allowed_states  = self.allowed_states
        
        if tags:
            maxTags = self.max_tags
            
            if(len(tags)>maxTags):
                weightTab = {}
                for tag in tags:
                    query = {}
                    if allowed_states:
                        query['review_state'] = allowed_states
                    query['portal_type'] = friendly_types
                    weightTab[tag] = len(catalog.searchResults(Subject=tag, **query))
                    
                orderedTags = []
                for tag in weightTab.keys():
                    added=0
                    for i in range(0, len(orderedTags)):
                        if (weightTab[tag] > weightTab[orderedTags[i]]):
                            orderedTags.insert(i , tag)
                            added=1
                            break;
                    if (added==0):
                        orderedTags.append(tag)  
                           
                tags = orderedTags[:maxTags]
                        
            tags.sort(lambda x, y: cmp(x.lower(), y.lower()))
            return tuple(tags)
    
    def getResearchTags(self, REQUEST=None):
        """ Return tuple for the Portlet in Research Mode  """
        REQUEST = REQUEST or self.REQUEST
        siteTags = self.getSiteKeywords()
        subjects = REQUEST.get('subjects',[])
        if isinstance(subjects,str):
            subjects = [subjects]
           
        currentDocuments = self.getDocumentsForSubjects(subjects)
        otherSubjects = self.getOtherSubjects(currentDocuments, subjects)
        
        portletTags = subjects + otherSubjects
        portletTags.sort(lambda x, y: cmp(x.lower(), y.lower()))
        
        resultTags = []
        for tag in portletTags:
            
            if tag in subjects:
                tag_css_class = 'pageTag'
            else:
                tag_css_class = 'nonPageTag'
                
            size = self.tce_getTagSize(tag, 
                                    self.min_tags_size, 
                                    self.max_tags_size, 
                                    self.getScaledSizes(),
                                    siteTags,
                                    self.scaling_method)
            if size:
                resultTags.append( (tag, 
                                tag_css_class,
                                size
                                ))
        return tuple(resultTags)
    
    
    
    def getPortletTags(self, page_object, full_list=0):
        """Return tuple [ (name, class, size) ] for the portlet """
        i =0
        siteTags = self.getSiteKeywords()
        pageTags = self.getPageKeywords(page_object)

        if siteTags:
            maxTags = self.getMaxTags()
            
            if not pageTags:
                pageTags=[]
            
            portletTags = []
    
            for tag in siteTags:
                
                i +=1
                if ( (i > maxTags) and (full_list != 1) ):
                    break
                
                if tag in pageTags:
                    tag_css_class = 'pageTag'
                else:
                    tag_css_class = 'nonPageTag'
                
                size = self.tce_getTagSize(tag, 
                                          self.min_tags_size, 
                                          self.max_tags_size, 
                                          self.getScaledSizes(),
                                          siteTags,
                                          self.scaling_method)
                if size:
                    portletTags.append( (tag,
                                     tag_css_class,
                                     size,
                                     ))  
            return tuple(portletTags)
    
    def getMaxTags(self):
        """ Return the max number of tags allowed to display """
        return self.max_tags
     
    def getTagUrl(self, tag, REQUEST=None):
        """ Return the portlet's tags url to search page """
        REQUEST = REQUEST or self.REQUEST
        request_subjects = REQUEST.form.get('subjects',[])
        if isinstance(request_subjects, str):
            request_subjects = [request_subjects]
                
        if tag in request_subjects:
            index_tag = request_subjects.index(tag)
            subjects = request_subjects[:index_tag] + request_subjects[index_tag+1:] 
        else:
            subjects = request_subjects + [tag]
        
        if len(subjects) == 0:
                return '/tag_cloud_explorer'
        
        link = '/tag_cloud_explorer_results'
        i = 0
        for subject in subjects:
            i += 1
            if i == 1:
                link += '?'
            else:
                link += '&'
            link += "subjects:list=" + subject  
                 
        return link
    
    def getExploreStatus(self, REQUEST=None):
        """ Return the status of the portlet : 1 for tagexplorer , none for tagcloud """
        REQUEST = REQUEST or self.REQUEST
        subjects = REQUEST.get('subjects', [])
        if subjects:
            return 1
    
    def getDocumentsForResearch(self, subjects):
        """ Return pointer to documents """
        
        catalog = getToolByName(self, 'portal_catalog')
        documents=[]
        
        query = {}
        if self.allowed_states:
            query['review_state'] = self.allowed_states
        query['sort_on'] = 'portal_type'
        query['sort_order'] = 'reverse'
            
        if subjects:
            return catalog.searchResults(
                               Subject = {'query':subjects,'operator':'and'},
                               **query)
           
    def getDocumentsForSubjects(self, subjects):
        """ Return 'self made' documents associated to subjects """
        
        catalog = getToolByName(self, 'portal_catalog')
        documents=[]
        query = {}
        if self.allowed_states:
            query['review_state'] = self.allowed_states
        query['sort_on'] = 'portal_type'
        query['sort_order'] = 'reverse'

        if subjects:
            for o in catalog.searchResults(
                                       Subject = {'query':subjects,'operator':'and'}
                                     , **query):
                url=o.getURL()
                title= o.Title or o.getId           
                
                documents.append( {'title':title
                                 , 'url':url
                                 , 'icon':o.getIcon
                                 , 'subject':o.Subject
                                 , 'author':o.Creator
                                 , 'modified':o.ModificationDate
                                 , 'description':o.Description} )
        return documents
    
    def getOtherSubjects(self, documents, subjects):
        """ Return other selectable subjects for documents """
        
        other_subjects= []
        if documents:
            for d in documents:
                for s in d['subject']:
                    if (s in subjects) or (s in other_subjects):
                        pass
                    else:
                        other_subjects += [s]       
        return other_subjects
            
        
InitializeClass(TagCloudExplorerTool)
